using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Microsoft.Extensions.Logging;
using IceCreamShopContracts.BindingModels;
using IceCreamShopContracts.BusinessLogicsContracts;
using IceCreamShopContracts.SearchModels;
using IceCreamShopContracts.StoragesContracts;
using IceCreamShopContracts.ViewModels;
using AbstractIceCreamShopDataModels.Enums;
using IceCreamBusinessLogic.MailWorker;
using AbstractIceCreamShopDataModels.Models;

namespace IceCreamBusinessLogic.BusinessLogics
{
    public class OrderLogic : IOrderLogic
    {
        private readonly ILogger _logger;
        private readonly IOrderStorage _orderStorage;
		private readonly AbstractMailWorker _mailWorker;
		private readonly IClientLogic _clientLogic;
        private readonly IShopStorage _shopStorage;
        private readonly IShopLogic _shopLogic;
        private readonly IIceCreamStorage _iceCreamStorage;

		public OrderLogic(ILogger<OrderLogic> logger, IOrderStorage orderStorage, IIceCreamStorage icecreamStorage, IShopLogic shopLogic, IClientLogic clientLogic, AbstractMailWorker mailWorker)
		{
			_logger = logger;
			_shopLogic = shopLogic;
			_iceCreamStorage = icecreamStorage;
			_orderStorage = orderStorage;
			_mailWorker = mailWorker;
			_clientLogic = clientLogic;
		}

		public bool CreateOrder(OrderBindingModel model)
        {
            CheckModel(model);
            if(model.Status != OrderStatus.Неизвестен)
            {
                _logger.LogWarning("Insert operation failed. Order status incorrect.");
                return false;
            }
            model.Status = OrderStatus.Принят;
			var result = _orderStorage.Insert(model);
			if (result == null)
			{
				_logger.LogWarning("Insert operation failed");
				return false;
			}
			SendOrderStatusMail(result.ClientId, $"Новый заказ создан. Номер заказа #{result.Id}", $"Заказ #{result.Id} от {result.DateCreate} на сумму {result.Sum:0.00} принят");
			return true;
		}

        public bool DeliveryOrder(OrderBindingModel model)
        {
            return SetNewStatus(model, OrderStatus.Готов);
        }

        public bool FinishOrder(OrderBindingModel model)
        {
            return SetNewStatus(model, OrderStatus.Выдан);
        }

        public bool TakeOrderInWork(OrderBindingModel model)
        {
            return SetNewStatus(model, OrderStatus.Выполняется);
        }

        public List<OrderViewModel>? ReadList(OrderSearchModel? model)
        {
            _logger.LogInformation("ReadList. OrderID:{Id}", model?.Id);
            var list = model == null ? _orderStorage.GetFullList() : _orderStorage.GetFilteredList(model);
            if (list == null)
            {
                _logger.LogWarning("ReadList return null list");
                return null;
            }
            _logger.LogInformation("ReadList. Count:{Count}", list.Count);
            return list;
        }

        private void CheckModel(OrderBindingModel model)
        {
            if (model == null)
            {
                throw new ArgumentNullException(nameof(model));
            }
            if (model.Id < 0)
            {
                throw new ArgumentNullException("Некорректный идентификатор заказа", nameof(model.Id));
            }
            if (model.IceCreamId < 0)
            {
                throw new ArgumentNullException("Некорректный идентификатор мороженого", nameof(model.IceCreamId));
            }
            if (model.Count <= 0)
            {
                throw new ArgumentNullException("Количество мороженого в заказе должно быть больше 0", nameof(model.Count));
            }
            if (model.Sum <= 0)
            {
                throw new ArgumentNullException("Сумма заказа должна быть больше 0", nameof(model.Sum));
            }
            _logger.LogInformation("Order. OrderID:{Id}.Sum:{ Sum}. DocumentId: { DocumentId}", model.Id, model.Sum, model.IceCreamId);
        }

		public bool SetNewStatus(OrderBindingModel rawModel, OrderStatus newStatus)
		{
			var viewModel = _orderStorage.GetElement(new OrderSearchModel
			{
				Id = rawModel.Id
			});

			if (viewModel == null)
			{
				_logger.LogWarning("Order model not found");
				return false;
			}

			OrderBindingModel model = new OrderBindingModel
			{
				Id = viewModel.Id,
				IceCreamId = viewModel.IceCreamId,
				Status = viewModel.Status,
				DateCreate = viewModel.DateCreate,
				DateImplement = viewModel.DateImplement,
				Count = viewModel.Count,
				Sum = viewModel.Sum,
				ImplementerId = viewModel.ImplementerId
			};
			if (rawModel.ImplementerId.HasValue)
			{
				model.ImplementerId = rawModel.ImplementerId;
			}

			CheckModel(model);
			if (model.Status + 1 != newStatus && model.Status != OrderStatus.Ожидается)
			{
				_logger.LogWarning("Status update to " + newStatus.ToString() + " operation failed. Order status incorrect.");
				return false;
			}

			if (newStatus == OrderStatus.Готов)
			{
				var icecream = _iceCreamStorage.GetElement(new IceCreamSearchModel() { Id = model.IceCreamId });
				if (icecream == null)
				{
					_logger.LogWarning("Status update to " + newStatus.ToString() + " operation failed. Icecream not found.");
					return false;
				}
				if (CheckSupply(icecream, model.Count) == false)
				{
					_logger.LogWarning("Status update to " + newStatus.ToString() + " operation failed. Shop supply error.");
					model.Status = OrderStatus.Ожидается;
					_orderStorage.Update(model);
					return false;
				}
			}

			model.Status = newStatus;
			if (model.Status == OrderStatus.Выдан) model.DateImplement = DateTime.Now;
			var result = _orderStorage.Update(model);
			if (result == null)
			{
				model.Status--;
				_logger.LogWarning("Update operation failed");
				return false;
			}
			SendOrderStatusMail(result.ClientId, $"Изменен статус заказа #{result.Id}", $"Заказ #{result.Id} изменен статус на {result.Status}");
			return true;
		}

		public bool CheckSupply(IIceCreamModel model, int count)
		{
			if (count <= 0)
			{
				_logger.LogWarning("Check then supply operation error. icecream count < 0.");
				return false;
			}

			int freeSpace = 0;
			foreach (var shop in _shopStorage.GetFullList())
			{
				freeSpace += shop.IceCreamMaxCount;
				foreach (var icecream in shop.ShopIceCreams)
				{
					freeSpace -= icecream.Value.Item2;
				}
			}

			if (freeSpace - count < 0)
			{
				_logger.LogWarning("Check then supply operation error. There's no place for new icecreams in shops.");
				return false;
			}

			foreach (var shop in _shopStorage.GetFullList())
			{
				freeSpace = shop.IceCreamMaxCount;
				foreach (var icecream in shop.ShopIceCreams)
				{
					freeSpace -= icecream.Value.Item2;
				}
				if (freeSpace == 0)
				{
					continue;
				}
				if (freeSpace - count >= 0)
				{
					if (_shopLogic.SupplyIceCreams(new() { Id = shop.Id }, model, count)) count = 0;
					else
					{
						_logger.LogWarning("Supply error");
						return false;
					}
				}
				if (freeSpace - count < 0)
				{
					if (_shopLogic.SupplyIceCreams(new() { Id = shop.Id }, model, freeSpace)) count -= freeSpace;
					else
					{
						_logger.LogWarning("Supply error");
						return false;
					}
				}
				if (count <= 0)
				{
					return true;
				}
			}
			return false;
		}

        public OrderViewModel? ReadElement(OrderSearchModel model)
        {
            if (model == null)
            {
                throw new ArgumentNullException(nameof(model));
            }
            _logger.LogInformation("ReadElement. Id:{ Id}", model.Id);
            var element = _orderStorage.GetElement(model);
            if (element == null)
            {
                _logger.LogWarning("ReadElement element not found");
                return null;
            }
            _logger.LogInformation("ReadElement find. Id:{Id}", element.Id);
            return element;
        }

		private bool SendOrderStatusMail(int clientId, string subject, string text)
		{
			var client = _clientLogic.ReadElement(new() { Id = clientId });
			if (client == null)
			{
				return false;
			}
			_mailWorker.MailSendAsync(new()
			{
				MailAddress = client.Email,
				Subject = subject,
				Text = text
			});
			return true;
		}
	}
}
